"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)
Date Created:   03/03/2019

Contains the classes for interfacing with training command group.
"""
import os
import click
import shutil
from datetime import datetime
from pathlib import Path
from colorama import Fore
from ravenml.utils.local_cache import RMLCache
from ravenml.utils.question import cli_spinner, user_input, user_selects, user_confirms
from ravenml.utils.dataset import get_dataset_names, get_dataset
from ravenml.data.interfaces import Dataset

class TrainInput(object):
    """Represents a training input. Contains all plugin-independent information
    necessary for training. Plugins can define their own behavior for getting
    additional information.

    Args:
        config (dict): training config dict
        plugin_name (str): name of the plugin that will use this TrainInput.
            Train command (commands.py) sets gives the exact plugin command (i.e, tf-bbox, etc)

    Attributes:
        config (dict): full training config dict as loaded from config yaml file
        plugin_cache (RMLCache): RMLCache for this plugin. Created at 
            ~/.ravenML/<plugin_name>
        artifact_path (Path): path to save artifacts. Points to temp/ inside
            the root of plugin_cache if uploading to S3, otherwise points
            to user defined local path.
        dataset (Dataset): Dataset object for this training run.
        metadata (dict): dictionary of metadata about this training.
            Automatically populated with common data, plugins add more as needed.
        plugin_metadata (dict): dictionary within full metadata dict where plugins
            place their specific metadata. Plugins should never assign to this
            attribute as it will break the relationship between plugin_metadata and metadata.
        plugin_config (dict): plugin section of config dict. Plugins look here
            for plugin-specific configuration.
    """
    def __init__(self, config:dict=None, plugin_name:str=None):
        """ Keyword args must be used for this class to work with the @pass_train pass decorator.
        That decorator is defined in options.py. It is applied to commands to ensure that a 
        TrainInput object exists in the click Context at ctx.obj. Click pass decorators 
        do this by creating an object with no constructor arguments at ctx.obj if one does
        not already exist.

        Therefore, this constructor is ONLY called with no args when a command decorated with 
        @pass_train is triggered and no TrainInput exists yet at ctx.obj. This behavior is used
        to know when the user has called a plugin training command but did not provide a config 
        filepath via --config.
        
        The train command defined at commands.py will ONLY create a TrainInput at ctx.obj
        if --config was used. This prevents undesired creation of TrainInput objects during
        execution non-training plugin commands, as click always executes the code in the train
        command on its way down the chain to the plugin command itself.
        
        tl:dr: if this constructor is called without keyword args set, we know a user
        has called a plugin's train command without providing a config, so we fail.
        """
        if config is None or plugin_name is None:
            raise click.exceptions.UsageError(('You must provide the --config option '
                'on `ravenml train` when using this plugin command.'))

        ## Store config
        self.config = config
        
        ## Set up Local Cache
        # TODO: maybe create the subdir here?
        # currently the cache_name subdir is only created IF the plugin places files there
        self.plugin_cache = RMLCache(plugin_name)
        
        ## Set up Artifact Path
        ap = config.get('artifact_path')
        if ap is None:
            self.plugin_cache.ensure_clean_subpath('temp')
            self.plugin_cache.ensure_subpath_exists('temp')
            self.artifact_path = Path(self.plugin_cache.path / 'temp')
        else:
            ap = Path(os.path.expanduser(ap))
            # check if local path contains data
            if os.path.exists(ap) and os.path.isdir(ap) and len(os.listdir(ap)) > 0:
                if config.get('overwrite_local') or user_confirms('Local artifact storage location contains old data. Overwrite?'):
                    shutil.rmtree(ap)
                else:
                    click.echo(Fore.RED + 'Training cancelled.')
                    click.get_current_context().exit() 
            # create directory, need exist_ok since we only delete
            # if directory contains files
            # TODO: protect against paths to actual files
            os.makedirs(ap, exist_ok=True)
            self.artifact_path = ap
        
        ## Set up Dataset
        # prompt for dataset if not provided
        dataset_name = config.get('dataset')
        if dataset_name is None:
            dataset_options = cli_spinner('No dataset provided. Finding datasets on S3...', get_dataset_names)
            dataset_name = user_selects('Choose dataset:', dataset_options)
        # download dataset and populate field
        try:
            self.dataset = cli_spinner(f'Downloading {dataset_name} from S3...', 
                get_dataset, dataset_name)
        except ValueError as e:
            hint = 'dataset name, no such dataset exists on S3'
            raise click.exceptions.BadParameter(dataset_name, param=dataset_name, param_hint=hint)
    
        ## Set up Basic Metadata
        # TODO: add environment description, git hash, etc
        self.metadata = config.get('metadata', {})
        # handle user defined metadata fields
        if not self.metadata.get('created_by'):
            self.metadata['created_by'] = user_input('Please enter your first and last name:')
        if not self.metadata.get('comments'):
            self.metadata['comments'] = user_input('Please enter descriptive comments about this training:')
        
        # handle automatic metadata fields
        self.metadata['date_started_at'] = datetime.utcnow().isoformat() + "Z"
        self.metadata['dataset_used'] = self.dataset.metadata
        
        ## Set up fields for plugin use
        # NOTE: plugins should overwrite the architecture field to something
        # more specific/useful since it is used to name the final uploaded model
        self.metadata[plugin_name] = {'architecture': plugin_name}
        # plugins should only ACCESS the plugin_metadata attibute and add items. They should
        # NEVER assign to the attribute as it will break the reference to the overall metadata dict
        self.plugin_metadata = self.metadata[plugin_name]
        if not config.get('plugin'):
            raise click.exceptions.BadParameter(config, param=config, param_hint='config, no "plugin" field. Config was')
        else:
            self.plugin_config = config.get('plugin') 
            
class TrainOutput(object):
    """Represents a training output. Plugin training command functions return this object
    so it can be processed by the `process_result` callback registed on the train command.
    This callback handles killing EC2 instances and uploading the trained model to S3.
    
    Note this object does not contain the metadata associated with the training run,
    although that is technically an output. This is because metadata is created 
    and populated with common data in the TrainInput __init__ that runs prior to training.
    The TrainInput object for the training is still at ctx.obj when training completes
    and can be accessed by the process_result callback, therefore metadata is
    not passed back via TrainOutput. Instead plugins directly modify the metadata
    in the TrainInput object during training, and it is then accessed for upload at the end.

    Args:
        model_path (Path): path to final exported model
        extra_files (list): list of Path objects to extra files associated with the training
            These are uploaded under the "extras" prefix in the models S3 bucket.
            They are tied to the final model via a unique UUID attached to each model at upload,
            Ex: model is uploaded with UUID "12345" at "models/12345.pb".
            Extras are then at: "extras/12345.pb"
            See _upload_result in commands.py for more details

    Attributes:
        model_path (Path): path to final exported model
        extra_files (list): list of Path objects to extra files associated with the training
    """
    def __init__(self, model_path: Path, extra_files: list):
        self.model_path = model_path
        self.extra_files = extra_files
    


"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   04/05/2019

Command group for setting ravenml configuration.
"""

import click
from pathlib import Path
from colorama import Fore
from ravenml.utils.question import user_input, user_confirms
from ravenml.utils.config import CONFIG_FIELDS, get_config, update_config

### OPTIONS ###
image_bucket_name_opt = click.option(
    '-i', '--image-bucket', type=str, is_eager=True,
    help='Image bucket name.')
    
dataset_bucket_name_opt = click.option(
    '-d', '--dataset-bucket', type=str, is_eager=True,
    help='Dataset bucket name.')

model_bucket_name_opt = click.option(
    '-m', '--model-bucket', type=str, is_eager=True,
    help='Model artifact destination bucket name.')


### COMMANDS ###
@click.group(help='Configuration commands.')
@click.pass_context
def config(ctx: click.Context):
    """Configuration command group.

    Args:
        ctx (Context): click context object
    """
    ctx.obj = {}
    ctx.obj['from_show'] = False     # flag to indicate if entering update from show

@config.command(help='Show current config.')
@click.pass_context
def show(ctx: click.Context):
    """Show current config.
    
    Args:
        ctx (Context): click context object
    """
    try:
        # try and load the configuration
        config = get_config()
        for key, value in config.items():
            print(Fore.GREEN + key + ': ' + Fore.WHITE + value)
    except FileNotFoundError:
        # thrown when no configuration file is found
        click.echo(Fore.RED + 'No configuration found.')
        if user_confirms('Would you like to make a new configuration?', default=True):
            ctx.obj['from_show'] = True
            ctx.invoke(update)
    except ValueError:
        # thrown when current configuration file is invalid
        click.echo(Fore.RED + 'Current configuration file is invalid.')
        if user_confirms('Would you like to fix it?', default=True):
            ctx.obj['from_show'] = True
            ctx.invoke(update)

@config.command(help='Update current config.')
@click.pass_context
@model_bucket_name_opt
@dataset_bucket_name_opt
@image_bucket_name_opt
def update(ctx: click.Context, image_bucket: str, dataset_bucket: str, model_bucket: str):
    """Update current config.
    
    Args:
        ctx (Context): click context object
        artifact_bucket (str): artifact bucket name. None if not in no-user mode
        dataset_bucket (str): dataset bucket name. None if not in no-user mode
    """
    config = {}
    load_result = 0
    try: 
        # try and load the configuration
        config = get_config()
    except FileNotFoundError:
        # thrown when no configuration file is found
        # checks to see if flag is set to indicate we arrived here from show
        if not ctx.obj['from_show']:
            click.echo(Fore.RED + 'No configuration found to update. A new one will be created.')
        load_result = 1
    except ValueError:
        # thrown current configuration file is invalid
        # checks to see if flag is set to indicate we arrived here from show
        if not ctx.obj['from_show']:
            click.echo(Fore.RED + 'Current configuration is invalid. A new one will be created.')
        load_result = 2

    # map options into dict
    user_options = {
        'image_bucket_name': image_bucket,
        'dataset_bucket_name': dataset_bucket,
        'model_bucket_name': model_bucket,
    }
    
    # if options were passed, just update those options
    if (image_bucket or dataset_bucket or model_bucket):
        # only use options if existing config was valid or a new config is beig created
        if load_result == 0 or load_result == 1:
            for field, value in user_options.items():
                if value:
                    config[field] = value
        else:
            click.echo(Fore.RED + "Passed options ignored since current configuration is invalid.")
    # otherwise iterate over each field
    else:
        for field in CONFIG_FIELDS:
            # only make field editing optional if existing config was loaded successfully
            if load_result == 0:
                if user_confirms('Edit ' + field + '?'):
                    config[field] = user_input(field + ':', default=config[field]) 
            else:
                config[field] = user_input(field + ':')

    # save updates
    update_config(config)



"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)
Date Created:   03/18/2019

Handles local file caching for ravenml.
"""

import os
import shutil
from pathlib import Path

# local cache root path for ravenml application
RAVENML_LOCAL_STORAGE_PATH = Path(os.path.expanduser('~/.ravenML'))

class RMLCache(object):
    """Represents a local storage cache. Provides functions for
    ensuring the cache exists and making subpaths within it.

    Args:
        path (Path): Absolute path in filesystem to serve as root of the local cache.
    
    Attributes:
        path (Path): Absolute path in filesystem to root of the local cache.
    """

    # local cache root path for ravenml application
    RAVENML_LOCAL_STORAGE_PATH = Path(os.path.expanduser('~/.ravenML'))
    
    def __init__(self, path:str='.'):
        self.path = RAVENML_LOCAL_STORAGE_PATH / Path(path)
        # self.ensure_exists()

    def ensure_exists(self):
        """Ensures that the local storage cache exists on the machine.
        """
        os.makedirs(self.path, exist_ok=True)

    def subpath_exists(self, subpath: str) -> bool:
        """Checks if a subpath within the local storage cache exists.
        
        Args:
            subpath (str): desired subpath (i.e 'datasets/my_dataset')
        
        Returns:
            bool: true if subpath exists, false if not
        """
        return os.path.exists(self.path / Path(subpath))

    def ensure_subpath_exists(self, subpath: str):
        """Ensures the subpath exists in the local storage cache.
        
        Args:
            subpath (str): desired subpath (i.e 'datasets/my_dataset')
        """
        os.makedirs(self.path / Path(subpath), exist_ok=True)
    
    def ensure_clean_subpath(self, subpath:str):
        if self.subpath_exists(subpath):
            shutil.rmtree(self.path / Path(subpath))

    def clean(self) -> bool:
        """Cleans local storage cache.
        
        Returns:
            bool: True if successfully removed cache, false if no cache found for removal
        """
        try:
            shutil.rmtree(self.path)
            return True
        except FileNotFoundError:
            return False
    


"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)
Date Created:   05/03/2019

Provides useful helper functions for training plugins.
"""

import click

def raise_parameter_error(option, hint: str):
    raise click.exceptions.BadParameter(option, param=option, param_hint=hint)

### DYNAMIC IMPORT FUNCTION ###
# NOTE: this function should be used in all plugins, but the function is NOT
# importable because of the use of globals(). You must copy the code.
# function is derived from https://stackoverflow.com/a/46878490

# def _dynamic_import(modulename, shortname = None, asfunction = False):
#     """ Function to dynamically import python modules into the global scope.

#     Args:
#         modulename (str): name of the module to import (ex: os, ex: os.path)
#         shortname (str, optional): desired shortname binding of the module (ex: import tensorflow as tf)
#         asfunction (bool, optional): whether the shortname is a module function or not (ex: from time import time)
        
#     Examples:
#         Whole module import: i.e, replace "import tensorflow"
#         >>> _dynamic_import('tensorflow')
        
#         Named module import: i.e, replace "import tensorflow as tf"
#         >>> _dynamic_import('tensorflow', 'tf')
        
#         Submodule import: i.e, replace "from object_detction import model_lib"
#         >>> _dynamic_import('object_detection.model_lib', 'model_lib')
        
#         Function import: i.e, replace "from ravenml.utils.config import get_config"
#         >>> _dynamic_import('ravenml.utils.config', 'get_config', asfunction=True)
        
#     """
#     if shortname is None: 
#         shortname = modulename
#     if asfunction is False:
#         globals()[shortname] = importlib.import_module(modulename)
#     else:        
#         globals()[shortname] = getattr(importlib.import_module(modulename), shortname)




"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   02/23/2019

Command group for training in ravenml.
"""

import click
import json
import shortuuid
import boto3
import yaml
from urllib.request import urlopen
from urllib.error import URLError
from pathlib import Path
from pkg_resources import iter_entry_points
from click_plugins import with_plugins
from ravenml.train.interfaces import TrainInput, TrainOutput
from ravenml.utils.question import cli_spinner
from ravenml.utils.aws import upload_file_to_s3, upload_dict_to_s3_as_json

EC2_INSTANCE_ID_URL = 'http://169.254.169.254/latest/meta-data/instance-id'

### OPTIONS ###
config_opt = click.option(
    '-c', '--config', type=str, help='Path to config file. Defaults to ~/ravenML_configs/config.yaml'
)

### COMMANDS ###
@with_plugins(iter_entry_points('ravenml.plugins.train'))
@click.group(help='Training commands.')
@click.pass_context
@config_opt
def train(ctx: click.Context, config: str):
    """ Training command group.
    
    Args:
        ctx (Context): click context object
        config (str): Path to config yaml file for this training run. Required
            when a user is calling a plugin command decorated with @pass_train
    """
    # check if config flag was passed, if not simply carry on to child command
    if config:
        # attempt to load config
        try:
            with open(Path(config), 'r') as stream:
                train_config = yaml.safe_load(stream)
        except Exception as e:
            hint = 'config, no such file exists'
            raise click.exceptions.BadParameter(config, ctx=ctx, param=config, param_hint=hint)
        # trigger TrainInput creation, note this may prompt the user depending on the config file used
        ctx.obj = TrainInput(train_config, ctx.invoked_subcommand)

@train.resultcallback()
@click.pass_context
def process_result(ctx: click.Context, result: TrainOutput, config: str):
    """Processes the result of a training by analyzing the given TrainOutput object.
    This callback is called after ANY command originating from the train command 
    group, hence the check to see if a result was actually returned - plugins
    simply do not return a TrainOutput from non-training commands.

    Args:
        ctx (Context): click context object
        ti (TrainInput): TrainInput object stored at ctx.obj, created at start of training.
            This object contains the metadata from the run and the path to the
            training artifacts (at ti.artifact_path) which is printed to the user
            after a local training.
        result (TrainOutput): training output object returned by training plugin
        config (str): config option from train command. Click requires that command
            callbacks accept the options from the original command.
    """
    if result is not None:
        # only plugin training commands that return a TrainOutput will activate this block
        # thus ctx.obj will always be a TrainInput object
        # NOTE: you cannot use the @pass_train decorator on process_result, otherwise on
        # non-training plugin commands, the TrainInput __init__ will be called by Click
        # when process_result runs and no TrainInput is at ctx.obj
        ti = ctx.obj    
        # upload if not in local mode, determined by user defined artifact_path field in config
        if not ti.config.get('artifact_path'):
            uuid = cli_spinner('Uploading artifacts...', _upload_result, result, ti.metadata)
            click.echo(f'Artifact UUID: {uuid}')
        else:
            with open(ti.artifact_path / 'metadata.json', 'w') as f:
                json.dump(ti.metadata, f, indent=2)
            click.echo(f'LOCAL MODE: Not uploading model to S3. Model is located at: {ti.artifact_path}')
            
        # stop, terminate, or do nothing to ec2 based on policy
        ec2_policy = ti.config.get('ec2_policy')
        # check if the policy is to stop or terminate
        if ec2_policy == None or ec2_policy == 'stop' or ec2_policy == 'terminate':
            policy_str = ec2_policy if ec2_policy else 'default'
            click.echo(f'Checking for EC2 instance and applying policy "{policy_str}"...')
            try:
                # grab ec2 id
                with urlopen(EC2_INSTANCE_ID_URL, timeout=5) as url:
                    ec2_instance_id = url.read().decode('utf-8')
                click.echo(f'EC2 Runtime detected.')
                client = boto3.client('ec2')
                # default is stop
                if ec2_policy == None or ec2_policy == 'stop':
                    click.echo("Stopping...")
                    client.stop_instances(InstanceIds=[ec2_instance_id], DryRun=False)
                else:
                    click.echo("Terminating...")
                    client.terminate_instances(InstanceIds=[ec2_instance_id], DryRun=False)
            except URLError:
                click.echo('No EC2 runtime detected. Doing nothing.')
        else:
            click.echo('Not checking for EC2 runtime since policy is to keep running.')
    return result


### HELPERS ###
def _upload_result(result: TrainOutput, metadata: dict):
    """ Wraps upload procedure into single function for use with cli_spinner.

    Generates a UUID for the model and uploads all artifacts.

    Args:
        result (TrainOutput): TrainOutput object, to be uploaded
        metadata (dict): metadata associated with this run, to be uploaded
    
    Returns:
        str: uuid assigned to result on upload
    """
    shortuuid.set_alphabet('23456789abcdefghijkmnopqrstuvwxyz')
    uuid = shortuuid.uuid()
    model_name = f'{metadata["architecture"]}_{uuid}.pb'
    upload_file_to_s3('models', result.model_path, alternate_name=model_name)
    upload_dict_to_s3_as_json(f'models/metadata_{uuid}', metadata)
    if result.extra_files != []:
        for fp in result.extra_files:
            upload_file_to_s3(f'extras/{uuid}', fp)
    return uuid



from setuptools import setup, find_packages

setup(
    name='ravenml',
    version='1.1',
    description='ML Training CLI Tool',
    license='MIT',
    packages=find_packages(),
    author='Carson Schubert, Abhi Dhir, Pratyush Singh',
    author_email='carson.schubert14@gmail.com',
    keywords= ['machine learning', 'data science'],
    download_url = 'https://github.com/autognc/ravenML/archive/v1.1.tar.gz',
    install_requires=[
        'Click>=7.0',
        'click-plugins>=1.0.4',
        'questionary>=1.0.2',
        'boto3>=1.9.86', 
        'shortuuid>=0.5.0',
        'halo>=0.0.26'
        'colorama>=0.3.9',
        'pyaml>=19.4.1',
    ],
    tests_require=[
        'pytest',
        'moto'
    ],
    entry_points='''
      [console_scripts]
      ravenml=ravenml.cli:cli
    ''',
)
      